// Copyright (C) 2017-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package neo
// master node

// Master organization
//
// Master is organized as follows:
//
// - main task that controls whole logic of master working. It spawns
//   subtasks to implement that logic and communicate with the subtask via channels. XXX
//   Main is the only mutator of nodeTab, partTab, etc.
//
// - accept task that accepts incoming connections and hands them over to main
//   via nodeComeq.
//
// - per peer workers are spawned that interact with main via channels
// - δnodeTab, δpartTab updates are proxied to peer by another per-peer task
//
// XXX
//
// XXX
// master manages node and partition tables and broadcast their updates
// to all connected nodes.

import (
	"context"
	stderrors "errors"
	"fmt"
	"sync"
	"time"

	xxcontext "lab.nexedi.com/kirr/go123/xcontext"
	"lab.nexedi.com/kirr/go123/xnet"
	"lab.nexedi.com/kirr/go123/xsync"

	"lab.nexedi.com/kirr/neo/go/neo/internal/xtime"
	"lab.nexedi.com/kirr/neo/go/neo/neonet"
	"lab.nexedi.com/kirr/neo/go/neo/proto"
	"lab.nexedi.com/kirr/neo/go/neo/xneo"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/internal/log"
	"lab.nexedi.com/kirr/neo/go/internal/task"
	"lab.nexedi.com/kirr/neo/go/internal/xcontext"
	"lab.nexedi.com/kirr/neo/go/internal/xio"
)

// Master is a node overseeing and managing how whole NEO cluster works.
type Master struct {
	node *xneo.Node

	// whole Run runs under runCtx
	runCtx context.Context

	// "global" workgroup under which main, accept and tasks, that should
	// last for whole Run time, are spawned.
	mainWG *xsync.WorkGroup

	// main <- node come or go
	nodeComeq    chan nodeCome	// main <- accept "node connected"
	nodeLeaveq   chan nodeLeave	// main <- peerWG.wait "node (should be) disconnected"

	// in addition to nodeTab (which keeps information about a node) tasks
	// and mastering context that are specific to a peer are organized
	// around peerTab[peer.nid].
	peerTab map[proto.NodeID]*_MasteredPeer


	// last allocated oid & tid
	// XXX how to start allocating oid from 0, not 1 ?
	// TODO mu
	lastOid zodb.Oid
	lastTid zodb.Tid

	// channels controlling main driver
	ctlStart    chan chan error	// request to start cluster
	ctlStop     chan chan struct{}	// request to stop  cluster
	ctlShutdown chan chan error	// request to shutdown cluster XXX with ctx ?

	// so tests could override
	monotime func() float64
}

// nodeCome represents "node connects" event.
// XXX main <- accept
type nodeCome struct {
	req    *neonet.Request
	idReq  *proto.RequestIdentification // we received this identification request
}

// nodeLeave represents "node (should be) disconnected" event.
type nodeLeave struct {
	peer *_MasteredPeer
	err  error // XXX needed ?
}

// _MasteredPeer represents context for all tasks related to one peer driven by master.
//
// .notify
// .wait (run under mainWG)
type _MasteredPeer struct {
	node *xneo.PeerNode

	// all tasks are spawned under wg. If any task fails - whole wg is canceled.
	// cancel signals all tasks under wg to stop.
	wg     *xsync.WorkGroup
	cancel func()

	// snapshot of nodeTab/partTab/stateCode when peer was accepted by main.
	state0 *xneo.ClusterStateSnapshot
	// main -> peerWG.notify  δnodeTab/δpartTab/δstateCode.
	notifyq chan _ΔClusterState
	// notifyqOverflow becomes ready if main detects that peer is to slow to consume updates
	// XXX no need? (peer.notify is canceled via peerWork.cancel)
	notifyqOverflow chan struct{}

	acceptDone chan struct{} // ready after initial accept sequence is sent to the peer
}

// _ΔClusterState represents δnodeTab/δpartTab/δClusterState.
type _ΔClusterState interface { δClusterState() }
type _ΔNodeTab struct {
	proto.NodeInfo // new value for change of 1 nodeTab entry
}
type _ΔPartTab struct {
	// XXX
}
type _ΔStateCode struct {
	proto.ClusterState // new value
}

func (_ *_ΔNodeTab)   δClusterState() {}
func (_ *_ΔPartTab)   δClusterState() {}
func (_ *_ΔStateCode) δClusterState() {}


// NewMaster creates new master node.
//
// Use Run to actually start running the node.
func NewMaster(clusterName string, net xnet.Networker) *Master {
	m := &Master{
		node: xneo.NewNode(proto.MASTER, clusterName, net, ""),

		ctlStart:	make(chan chan error),
		ctlStop:	make(chan chan struct{}),
		ctlShutdown:	make(chan chan error),

		nodeComeq:	make(chan nodeCome),
		nodeLeaveq:	make(chan nodeLeave),

		peerTab:        make(map[proto.NodeID]*_MasteredPeer),

		monotime:	xtime.Mono,
	}

	return m
}

// Start requests cluster to eventually transition into running state.
//
// It returns an error if such transition is not currently possible to begin
// (e.g. partition table is not operational).
// It returns nil if the transition began.
//
// NOTE upon successful return cluster is not yet in running state - the transition will
// take time and could be also automatically aborted due to cluster environment change (e.g.
// a storage node goes down).
func (m *Master) Start() error {
	ech := make(chan error)
	m.ctlStart <- ech
	return <-ech
}

// Stop requests cluster to eventually transition into recovery state.
func (m *Master) Stop()  {
	ech := make(chan struct{})
	m.ctlStop <- ech
	<-ech
}

// Shutdown requests all known nodes in the cluster to stop.
// XXX + master's run to finish ?
func (m *Master) Shutdown() error {
	panic("TODO")
}


// setClusterState sets .clusterState and notifies subscribers.
func (m *Master) setClusterState(state proto.ClusterState) {
	m.node.State.Code.Set(state)

	// TODO notify subscribers
	// <- _ΔStateCode{state}
}


// Run starts master node and runs it until ctx is cancelled or fatal error.
//
// The master will be serving incoming connections on l.
func (m *Master) Run(ctx context.Context, l xnet.Listener) (err error) {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel() // so that .runCtx is canceled if we return due to an error
		       // XXX ^^^ not needed - we first must wait for all spawned subtasks

	mynid := m.allocNID(proto.MASTER)

	addr := l.Addr()
	defer task.Runningf(&ctx, "%s", mynid)(&err)
	log.Infof(ctx, "listening on %s ...", addr)
	m.runCtx = ctx
	m.mainWG = xsync.NewWorkGroup(m.runCtx)


	// update our master & serving address in node
	naddr, err := proto.Addr(addr)
	if err != nil {
		return err
	}
	m.node.MasterAddr = addr.String()
	m.node.MyInfo = proto.NodeInfo{
		Type:	proto.MASTER,
		Addr:	naddr,
		NID:	mynid,
		State:	proto.RUNNING,
		IdTime:	proto.IdTimeNone,	// XXX ok?
	}

	// update nodeTab with self
	m.updateNodeTab(ctx, m.node.MyInfo)

	// wrap listener with link / identificaton hello checker
	lli := xneo.NewListener(neonet.NewLinkListener(l))

	// accept: accept incoming connections and pass them to main driver
	m.mainWG.Go(func(ctx context.Context) (err error) {
		defer task.Running(&ctx, "accept")(&err)

		// XXX dup in storage
		for {
			if ctx.Err() != nil {
				return ctx.Err()
			}

			req, idReq, err := lli.Accept(ctx)
			if err != nil {
				if !xcontext.Canceled(err) {
					log.Error(ctx, err)	// XXX throttle?
				}
				continue
			}

			// for storages the only incoming connection is for RequestIdentification
			// and then master only drives it. So close accept as noone will be
			// listening for it on our side anymore.
			switch idReq.NodeType {
			case proto.CLIENT:
				// ok

			case proto.STORAGE:
				fallthrough
			default:
				req.Link().CloseAccept()
			}

			// handover to main driver
			select {
			case m.nodeComeq <- nodeCome{req, idReq}:
				// ok

			case <-ctx.Done():
				// shutdown
				xio.LClose(ctx, req.Link())
				continue
			}
		}
	})

	// main: main driving logic
	m.mainWG.Go(m.main)

	err = m.mainWG.Wait()
	// change `... canceled` to just canceled?
	// (e.g. `master: accept: canceled` or `master: main: canceled` -> `master: canceled`)
	if ctx.Err() != nil {
		err = ctx.Err()
	}

	xio.LClose(ctx, lli)	// XXX here ok? (probbly not)

	return err
}

// main is the process that implements main master cluster management logic: node tracking, cluster
// state updates, scheduling data movement between storage nodes, etc.
func (m *Master) main(ctx context.Context) (err error) {
	defer task.Running(&ctx, "main")(&err)

	// NOTE Run's goroutine is the only mutator of nodeTab, partTab and other cluster state

	// XXX however since clients request state reading we should use node.StateMu?
	// XXX -> better rework protocol so that master pushes itself (not
	//     being pulled) to clients everything they need.
	//     -> it was reworked (see bf240897)

	for ctx.Err() == nil {
		// recover partition table from storages and wait till enough
		// storages connects us so that we can see the partition table
		// can be operational.
		//
		// Successful recovery means all ^^^ preconditions are met and
		// a command came to us to start the cluster.
		err := m.recovery(ctx)
		if err != nil {
			//log.Error(ctx, err)
			return err // recovery cancelled
		}

		// make sure transactions on storages are properly finished, in
		// case previously it was unclean shutdown.
		err = m.verify(ctx)
		if err != nil {
			//log.Error(ctx, err)
			continue // -> recovery
		}

		// provide service as long as partition table stays operational
		err = m.serve(ctx)
		if err != nil {
			//log.Error(ctx, err)
			continue // -> recovery
		}

		// XXX shutdown request ?
	}

	return ctx.Err()
}


// Cluster Recovery
// ----------------
//
// - starts from potentially no storage nodes known
// - accept connections from storage nodes
// - retrieve and recover latest previously saved partition table from storages
// - monitor whether partition table becomes operational wrt currently up nodeset
// - if yes - finish recovering upon receiving "start" command		XXX or autostart
// - start is also allowed if storages connected and say there is no partition
//   table saved to them (empty new cluster case).

// storRecovery is result of 1 storage node passing recovery phase.
type storRecovery struct {
	stor    *_MasteredPeer
	partTab *xneo.PartitionTable
	err     error

	// XXX + backup_tid, truncate_tid ?

	// XXX naming  cont? continue? unlock? unpause?
	ack     chan struct{} // main -> storCtlRecovery  "thanks; please continue" XXX explain why
}

// recovery drives cluster during recovery phase.
//
// when recovery finishes error indicates:
// - nil:  recovery was ok and a command came for cluster to start
// - !nil: recovery was cancelled
func (m *Master) recovery(ctx context.Context) (err error) {
	defer task.Running(&ctx, "recovery")(&err)

	m.setClusterState(proto.ClusterRecovering)
	ctx, rcancel := context.WithCancel(ctx)
	defer rcancel()

	recoveryq := make(chan storRecovery)
	inprogress := 0 // in-progress stor recoveries
	wg := &sync.WaitGroup{}

	start := false // whether we were instructed to start
//trace:event traceMasterStartReady(m *Master, ready bool)
	readyToStart := false // whether cluster currently can be operational or not
	updateReadyToStart := func() {
		var ready bool
		if m.node.State.PartTab.PTid == 0 {
			// new cluster - allow startup if we have some storages passed
			// recovery and there is no in-progress recovery running
			nup := 0
			for _, stor := range m.node.State.NodeTab.StorageList() {
				if stor.State > proto.DOWN {
					nup++
				}
			}
			ready = (nup > 0 && inprogress == 0)

		} else {
			ready = m.node.State.PartTab.OperationalWith(m.node.State.NodeTab)	// XXX + node state
		}

		if readyToStart != ready {
			readyToStart = ready
			traceMasterStartReady(m, ready)
		}
	}

	// XXX set cluster state = RECOVERY
	// XXX down clients

	// goStorCtlRecovery spawns recovery task on a storage peer.
	goStorCtlRecovery := func(stor *_MasteredPeer) {
		inprogress++
		wg.Add(1)
		stor.wg.Go(func(peerCtx context.Context) error {
			defer wg.Done()
			ctx, cancel := xxcontext.Merge/*Cancel*/(ctx, peerCtx)
			defer cancel()

			var pt *xneo.PartitionTable
			err := stor.run(ctx, func() error {
				var err error
				pt, err = storCtlRecovery(ctx, stor)
				return err
			})

			ack := make(chan struct{})
			recoveryq <- storRecovery{stor: stor, partTab: pt, err: err, ack: ack}
			<-ack

			// canceled recovery does not mean we should down the storage node
			if xcontext.Canceled(err) {
				err = nil
			}
			return err
		})
	}

	// start recovery on all storages we are currently in touch with
	for _, peer := range m.peerTab {
		if peer.node.Type == proto.STORAGE {
			goStorCtlRecovery(peer)
		}
	}

loop:
	for !(inprogress == 0 && readyToStart && start) {
		select {
		case <-ctx.Done():
			err = ctx.Err()
			break loop

		// request to start the cluster - if ok we exit replying ok
		// if not ok - we just reply not ok
		case ech := <-m.ctlStart:
			if readyToStart {
				log.Infof(ctx, "start command - we are ready")
				// reply "ok to start" after whole recovery finishes

				// XXX ok? we want to retrieve all recovery information first?
				// XXX or initially S is in PENDING state and
				// transitions to RUNNING only after successful recovery?

				rcancel()
				defer func() {
					// XXX can situation change while we are shutting down?
					// XXX -> recheck logic with checking PT operational ^^^
					// XXX    (depending on storages state)
					ech <- nil
				}()
				break loop
			}

			log.Infof(ctx, "start command - err - we are not ready")
			ech <- fmt.Errorf("start: cluster is non-operational")

		case ech := <-m.ctlStop:
			close(ech) // ok; we are already recovering

		// peer (should be) disconnected
		case n := <-m.nodeLeaveq:
			m.disconnectPeer(ctx, n.peer)
			updateReadyToStart()

		// node comes in and asks to be identified
		case n := <-m.nodeComeq:
			peer, ok := m.identify(ctx, n,
				// XXX only accept:
				//	- S -> PENDING
				//	- M
			)
			if !ok {
				break
			}

			// S -> start recovery
			if peer.node.Type == proto.STORAGE {
				goStorCtlRecovery(peer)
			}

		// a storage node came through recovery - let's see whether
		// ptid ↑ and if so we should take partition table from there
		case r := <-recoveryq:
			close(r.ack) // for <-"node leave" to happen after <-recovery in case of err
			inprogress--

			if r.err != nil {
				log.Error(ctx, r.err)
			} else {
				// we are interested in latest partTab
				// NOTE during recovery no one must be subscribed to
				// partTab so it is ok to simply change whole m.partTab
				// TODO -> updatePartTab()
				if r.partTab.PTid > m.node.State.PartTab.PTid {
					m.node.State.PartTab = r.partTab
				}
			}

			updateReadyToStart()
		}
	}

	// wait all workers to finish (which should come without delay since it was cancelled)
	// XXX not good - some of the rest of the storages can fail in the
	// meantime and this will lead to partTab to become non-opertional.
	// XXX also: some of the recoveries could still _succeed_ (e.g.
	// successfuly recovery send was already queued to recoveryq but not
	// yet received) - this successful recovery could bring us newer
	// partTab fro which we should reconsider whether we have all needed
	// nodes up and running.
	done := make(chan struct{})
	go func() {
		wg.Wait()
		close(done)
	}()

loop2:
	for {
		select {
		case r := <-recoveryq:
			close(r.ack)
			log.Error(ctx, r.err)


		case <-done:
			break loop2
		}
	}

	if err != nil {
		return err
	}

	// recovery successful - we are starting

	// S PENDING -> RUNNING
	// XXX recheck logic is ok for when starting existing cluster
	for _, stor := range m.node.State.NodeTab.StorageList() {
		if stor.State == proto.PENDING {
			m.updateNodeState(ctx, stor, proto.RUNNING)
		}
	}

	// if we are starting for new cluster - create partition table
	if m.node.State.PartTab.PTid == 0 {
		// XXX -> m.nodeTab.StorageList(State > DOWN)
		storv := []*xneo.PeerNode{}
		for _, stor := range m.node.State.NodeTab.StorageList() {
			if stor.State > proto.DOWN {
				storv = append(storv, stor)
			}
		}
		m.node.State.PartTab = xneo.MakePartTab(1 /* XXX hardcoded */, storv)
		m.node.State.PartTab.PTid = 1
		log.Infof(ctx, "creating new partition table: %s", m.node.State.PartTab)
	}

	return nil
}

// storCtlRecovery drives a storage node during cluster recovering state.
// it retrieves various ids and partition table from as stored on the storage
func storCtlRecovery(ctx context.Context, stor *_MasteredPeer) (_ *xneo.PartitionTable, err error) {
	slink := stor.node.Link()
	defer task.Runningf(&ctx, "%s recovery", stor.node.NID)(&err)

	// XXX cancel on ctx
	// XXX close slink on err? (if yes -> xcontext.WithCloseOnErrCancel)

	recovery := proto.AnswerRecovery{}
	err = slink.Ask1(&proto.Recovery{}, &recovery)
	if err != nil {
		return nil, err
	}

	resp := proto.AnswerPartitionTable{}
	err = slink.Ask1(&proto.AskPartitionTable{}, &resp)
	if err != nil {
		return nil, err
	}

	// reconstruct partition table from response
	pt := xneo.PartTabFromDump(resp.PTid, resp.RowList) // TODO handle resp.NumReplicas
	return pt, nil
}


var errStopRequested   = stderrors.New("stop requested")
var errClusterDegraded = stderrors.New("cluster became non-operatonal")


// Cluster Verification (data recovery)
// ------------------------------------
//
// - starts with operational partition table
// - saves recovered partition table to all storages
// - asks all storages for partly finished transactions and decides cluster-wide
//   which such transactions need to be either finished or rolled back.
// - executes decided finishes and rollbacks on the storages.
// - retrieve last ids from storages along the way.
// - once we are done without losing too much storages in the process (so that
//   partition table is still operational) we are ready to enter servicing state.

// verify drives cluster during verification phase.
//
// when verify finishes error indicates:
// - nil:  verification completed ok; cluster is ready to enter running state
// - !nil: verification failed; cluster needs to be reset to recovery state
//
// prerequisite for start: .partTab is operational wrt .nodeTab
func (m *Master) verify(ctx context.Context) (err error) {
	defer task.Running(&ctx, "verify")(&err)

	m.setClusterState(proto.ClusterVerifying)
	ctx, vcancel := context.WithCancel(ctx)
	defer vcancel()

	verifyq := make(chan storVerify) // <- result of stor verify task
	inprogress := 0                  // in-progress verify tasks

	// NOTE we don't reset m.lastOid / m.lastTid to 0 in the beginning of verification
	//      XXX (= py), rationale=?

	// goStorCtlVerify spawns verify task on a storage peer.
	goStorCtlVerify := func(stor *_MasteredPeer) {
		inprogress++
		stor.wg.Go(func(peerCtx context.Context) error {
			var lastOid zodb.Oid
			var lastTid zodb.Tid
			err := stor.run(ctx, func() error {
				var err error
				lastOid, lastTid, err = storCtlVerify(ctx, stor, m.node.State.PartTab)
				return err
			})

			ack := make(chan struct{})
			verifyq <- storVerify{stor: stor, lastOid: lastOid, lastTid: lastTid, err: err, ack: ack}
			<-ack

			// canceled verify does not mean we should down the storage node
			if xcontext.Canceled(err) {
				err = nil
			}
			return err
		})
	}

	// start verification on all storages we are currently in touch with
	for _, peer := range m.peerTab {
		if peer.node.Type == proto.STORAGE {
			goStorCtlVerify(peer)
		}
	}

	// during stop: react only to task completion and node leaving
	ctxDone   := ctx.Done()
	ctlStart  := m.ctlStart
	ctlStop   := m.ctlStop
	nodeComeq := m.nodeComeq
	err = nil
	stop := func(stopErr error) {
		if err != nil {
			return
		}
		err = stopErr
		vcancel()
		ctxDone   = nil
		ctlStart  = nil
		ctlStop   = nil
		nodeComeq = nil
	}

	for inprogress > 0 {
		select {
		case <-ctxDone:
			stop(ctx.Err())

		case ech := <-ctlStart:
			ech <- nil // we are already starting

		case ech := <-ctlStop:
			close(ech) // ok
			stop(errStopRequested)

		// peer (should be) disconnected
		case n := <-m.nodeLeaveq:
			m.disconnectPeer(ctx, n.peer)

			// if cluster became non-operational - we cancel verification
			if !m.node.State.PartTab.OperationalWith(m.node.State.NodeTab) {
				// XXX ok to instantly cancel? or better
				// graceful shutdown in-flight verifications?
				stop(errClusterDegraded)
			}

		// node comes in and asks to be identified
		case n := <-nodeComeq:
			peer, ok := m.identify(ctx, n,
				// XXX only accept:
				//	- S -> known ? RUNNING : PENDING
				//	- M
			)
			if !ok {
				break
			}

			// S -> start verification on it too
			if peer.node.Type == proto.STORAGE {
				goStorCtlVerify(peer)
			}


		// a storage node came through verification - adjust our last{Oid,Tid} if ok
		// on error check - whether cluster became non-operational and stop verification if so
		//
		// FIXME actually implement logic to decide to finish/rollback transactions
		case v := <-verifyq:
			close(v.ack) // XXX explain why (see recovery)
			inprogress--

			if v.err != nil {
				log.Error(ctx, v.err)

				// check partTab is still operational
				// if not -> cancel to go back to recovery
				if !m.node.State.PartTab.OperationalWith(m.node.State.NodeTab) {
					stop(errClusterDegraded)
				}
			} else {
				if v.lastOid > m.lastOid {
					m.lastOid = v.lastOid
				}
				if v.lastTid > m.lastTid {
					m.lastTid = v.lastTid
				}
			}
		}
	}

	return err
}

// storVerify is result of a storage node passing verification phase.
type storVerify struct {
	stor    *_MasteredPeer
	lastOid zodb.Oid
	lastTid zodb.Tid
	err	error

	ack chan struct{}
}

// storCtlVerify drives a storage node during cluster verifying (= starting) state.
func storCtlVerify(ctx context.Context, stor *_MasteredPeer, pt *xneo.PartitionTable) (lastOid zodb.Oid, lastTid zodb.Tid, err error) {
	// XXX link.Close on err	-> = xcontext.WithCloseOnErrCancel
	// XXX cancel on ctx		-> = ^^^

	slink := stor.node.Link()
	defer task.Runningf(&ctx, "%s verify", stor.node.NID)(&err)

	lastOid = zodb.InvalidOid
	lastTid = zodb.InvalidTid

	// send just recovered parttab so storage saves it
	err = slink.Send1(&proto.SendPartitionTable{
		PTid:	     pt.PTid,
		NumReplicas: 0, // FIXME hardcoded
		RowList:     pt.Dump(),
	})
	if err != nil {
		return
	}

	locked := proto.AnswerLockedTransactions{}
	err = slink.Ask1(&proto.LockedTransactions{}, &locked)
	if err != nil {
		return
	}

	if len(locked.TidDict) > 0 {
		// TODO vvv
		err = fmt.Errorf("TODO: non-ø locked txns: %v", locked.TidDict)
		return
	}

	last := proto.AnswerLastIDs{}
	err = slink.Ask1(&proto.LastIDs{}, &last)
	if err != nil {
		return
	}

	return last.LastOid, last.LastTid, nil
}


// Cluster Running (operational service)
// -------------------------------------
//
// - starts with operational parttab and all present storage nodes consistently
//   either finished or rolled-back partly finished transactions.
// - monitor storages come & go and if parttab becomes non-operational leave to recovery.
// - provide service to clients while we are here.
//
// TODO also plan data movement on new storage nodes appearing

// serveDone is the error returned after serving peer is finished.
type serveDone struct {
	peer *_MasteredPeer
	err  error

	ack chan struct{}
}

// serve drives cluster during running state.
//
// TODO document error meanings on return
//
// prerequisite for start: .partTab is operational wrt .nodeTab and verification passed
func (m *Master) serve(ctx context.Context) (err error) {
	defer task.Running(&ctx, "serve")(&err)

	m.setClusterState(proto.ClusterRunning)
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	servedq := make(chan serveDone) // <- result of a serve task
	inprogress := 0                   // in-progress serve tasks

	// goServe spawns serve task for a peer.
	goServe := func(peer *_MasteredPeer) {
		inprogress++
		peer.wg.Go(func(peerCtx context.Context) error {
			ctx, cancel := xxcontext.Merge/*Cancel*/(ctx, peerCtx)
			defer cancel()

			switch peer.node.Type {
			case proto.STORAGE:
				err = storCtlServe(ctx, peer)

			case proto.CLIENT:
				err = m.serveClient(ctx, peer)

			// XXX ADMIN
			}

			ack := make(chan struct{})
			servedq <- serveDone{peer: peer, err: err, ack: ack}
			<-ack

			// canceled serve does not necessarily mean we should down the peer
			if xcontext.Canceled(err) {
				err = nil
			}
			return err
		})
	}

	// spawn peer serve driver (it should be only storages on entry here?)
	for _, peer := range m.peerTab {
		// XXX clients? other nodes?
		// XXX note PENDING - not adding to serve; ok?
		if peer.node.Type == proto.STORAGE && peer.node.State == proto.RUNNING {
			goServe(peer)
		}
	}

	// during stop: react only to task completion and node leaving
	ctxDone   := ctx.Done()
	ctlStart  := m.ctlStart
	ctlStop   := m.ctlStop
	nodeComeq := m.nodeComeq
	err = nil
	stop := func(stopErr error) {
		if err != nil {
			return
		}
		err = stopErr
		cancel()
		ctxDone   = nil
		ctlStart  = nil
		ctlStop   = nil
		nodeComeq = nil

		// XXX tell storages to stop
	}

	for inprogress > 0 {
		select {
		case <-ctxDone:
			stop(ctx.Err())

		case ech := <-ctlStart:
			ech <- nil // we are already started

		case ech := <-ctlStop:
			close(ech) // ok
			stop(fmt.Errorf("stop requested"))

		// peer (should be) disconnected
		case n := <-m.nodeLeaveq:
			m.disconnectPeer(ctx, n.peer)

			// if cluster became non-operational - cancel serve
			if !m.node.State.PartTab.OperationalWith(m.node.State.NodeTab) {
				stop(errClusterDegraded)
			}

		// new connection comes in -> start serving it
		case n := <-nodeComeq:
			peer, ok := m.identify(ctx, n, /* XXX accept everyone */)
			if !ok {
				break
			}

			goServe(peer)

		// serving a peer is done
		case d := <-servedq:
			close(d.ack) // XXX explain why (see recovery)
			inprogress--
		}
	}

	return err
}

// storCtlServe drives a storage node during cluster serve state
func storCtlServe(ctx context.Context, stor *_MasteredPeer) (err error) {
	defer task.Runningf(&ctx, "%s serve", stor.node.NID)(&err)
	slink := stor.node.Link()

	// XXX current neo/py does StartOperation / NotifyReady as separate
	//     sends, not exchange on the same conn. - py draftly fixed: see
	//     https://lab.nexedi.com/kirr/neo/commit/4eaaf186 for context.
	ready := proto.NotifyReady{}
	err = slink.Ask1(&proto.StartOperation{Backup: false}, &ready)
	//err = slink.Send1(&proto.StartOperation{Backup: false})
	//if err != nil {
	//	return err
	//}
	//req, err := slink.Recv1()
	//if err != nil {
	//	return err
	//}
	//req.Close()	XXX must be after req handling
	//switch msg := req.Msg.(type) {
	//case *proto.NotifyReady:
	//	// ok
	//case *proto.Error:
	//	return msg
	//default:
	//	return fmt.Errorf("unexpected message %T", msg)
	//}


	// now wait in a loop
	// XXX this should be also controlling transactions
	for {
		select {
		// XXX stub
		case <-time.After(1*time.Second):
			//println(".")

		case <-ctx.Done():
			// FIXME also send StopOperation
			return ctx.Err()
		}
	}
}

// serveClient serves incoming client link.
func (m *Master) serveClient(ctx context.Context, cli *_MasteredPeer) (err error) {
	defer task.Runningf(&ctx, "%s: serve client", cli.node.NID)(&err)
	clink := cli.node.Link()

//	wg, ctx := errgroup.WithContext(ctx)	// XXX -> sync.WorkGroup
	defer xio.CloseWhenDone(ctx, clink)()	// XXX -> cli.ResetLink?  (better not here)

/* XXX kill
	// M -> C notifications about cluster state
	wg.Go(func() error {
		return m.keepPeerUpdated(ctx, clink)
	})
*/

	// M <- C requests handler
//	wg.Go(func() error {
		for {
			req, err := clink.Recv1()
			if err != nil {
				return err
			}

			resp := m.serveClient1(ctx, req.Msg)
			err = req.Reply(resp)
			req.Close()
			if err != nil {
				return err
			}
		}
//	})
//
//	return wg.Wait()
}

// serveClient1 prepares response for 1 request from client.
func (m *Master) serveClient1(ctx context.Context, req proto.Msg) (resp proto.Msg) {
	switch req := req.(type) {
	case *proto.LastTransaction:
		// XXX lock
		return &proto.AnswerLastTransaction{m.lastTid}

	default:
		return &proto.Error{proto.PROTOCOL_ERROR, fmt.Sprintf("unexpected message %T", req)}
	}
}

// ----------------------------------------

// disconnectPeer resets link to the peer and sets its state to DOWN in nodeTab.
// other peers are notified with δnodeTab about it.
// XXX place=?
func (m *Master) disconnectPeer(ctx context.Context, peer *_MasteredPeer) {
	// XXX log?
	peer.node.ResetLink(ctx)
	delete(m.peerTab, peer.node.NID)
	m.updateNodeState(ctx, peer.node, proto.DOWN)
}

// updateNodeTab = .nodeTab.Update  +  send δnodeTab to all subscribers.
// XXX place
// called from main master process. XXX
func (m *Master) updateNodeTab(ctx context.Context, nodeInfo proto.NodeInfo) *xneo.PeerNode {
	node := m.node.State.NodeTab.Update(nodeInfo)

	event := &_ΔNodeTab{nodeInfo}
	// XXX locking
	for nid, peer := range m.peerTab {
		// TODO change limiting by buffer size to limiting by time -
		// - i.e. detach peer if event queue grows more than 30s of time.
		select {
		case peer.notifyq <- event:
			continue // ok
		default:
		}

		// XXX try to move logging into other place - so that we could remove ctx arg here
		log.Warningf(ctx, "peer %s is slow -> detaching it", nid)
		close(peer.notifyqOverflow)
		// TODO delete(m.peerTab, nid) -> p.cancel()
		// XXX what else?
		panic("TODO")
	}

	return node
}

// XXX place
// XXX doc
func (m *Master) updateNodeState(ctx context.Context, node *xneo.PeerNode, state proto.NodeState) {
	nodei := node.NodeInfo
	// XXX skip if .State == state ?
	nodei.State = state
	m.updateNodeTab(ctx, nodei)
}


// ----------------------------------------

// identify processes identification request of just connected node and either accepts or declines it.
//
// If node identification is accepted .nodeTab and .peerTab are updated and
// corresponding peer entry is returned. New task is spawned to reply with
// either accept or reject.
//
// XXX If the peer is accepted (run something after initial accept completes)
func (m *Master) identify(ctx context.Context, n nodeCome) (peer *_MasteredPeer, ok bool) {
	// XXX also verify ? :
	// - NodeType valid
	// - IdTime ?

	nid := n.idReq.NID
	nodeType := n.idReq.NodeType

	err := func() *proto.Error {
		if n.idReq.ClusterName != m.node.ClusterName {
			return &proto.Error{proto.PROTOCOL_ERROR, "cluster name mismatch"}
		}

		if nid == 0 || nid == proto.NID(n.idReq.NodeType, 0) /* XXX <- stub for "temp" check */ {
			nid = m.allocNID(nodeType)
		}
		// XXX nid < 0 (temporary) -> reallocate if conflict ?
		// XXX check nid matches NodeType

		node := m.node.State.NodeTab.Get(nid)
		if node != nil {
			// reject - nid is already occupied by someone else
			// XXX check also for down state - it could be the same node reconnecting
			return &proto.Error{proto.PROTOCOL_ERROR, fmt.Sprintf("nid %v already used by another node", nid)}
		}

		// accept only certain kind of nodes depending on .clusterState, e.g.
		// XXX ok to have this logic inside identify? (better provide from outside ?)
		switch nodeType {
		case proto.CLIENT:
			if m.node.State.Code != proto.ClusterRunning {
				return &proto.Error{proto.NOT_READY, "cluster not operational"}
			}

		case proto.STORAGE:
			// ok

		// TODO +master, admin
		default:
			return &proto.Error{proto.PROTOCOL_ERROR, fmt.Sprintf("not accepting node type %v", nodeType)}
		}

		return nil
	}()

	subj := fmt.Sprintf("identify: %s (%s)", n.req.Link().RemoteAddr(), n.idReq.NID)
	if err != nil {
		log.Infof(ctx, "%s: rejecting: %s", subj, err)
		m.mainWG.Go(func(ctx context.Context) error {
			// XXX close link on ctx cancel
			n.req.Reply(err)
			n.req.Link().Close()
			// XXX log err (if any)
			return nil // not to cancel main by a failing reject
		})
		return nil, false
	}

	log.Infof(ctx, "%s: accepting as %s", subj, nid)

	accept := &proto.AcceptIdentification{
			NodeType:	proto.MASTER,
			MyNID:		m.node.MyInfo.NID,
			YourNID:	nid,
		}

	// update nodeTab
	var nodeState proto.NodeState
	switch nodeType {
	case proto.STORAGE:
		// FIXME py sets to RUNNING/PENDING depending on cluster state
		nodeState = proto.PENDING

	default:
		nodeState = proto.RUNNING
	}

	nodeInfo := proto.NodeInfo{
		Type:	nodeType,
		Addr:	n.idReq.Address,
		NID:	nid,
		State:	nodeState,
		IdTime:	proto.IdTime(m.monotime()),
	}

	node := m.updateNodeTab(ctx, nodeInfo)
	node.SetLink(n.req.Link())


	// create peer with nodeTab/partTab snapshot to push to accepted node
	// and subscribe it for updates.
	peerCtx, peerCancel := context.WithCancel(m.runCtx)
	// XXX add accept.NID to peerCtx task?
	peer = &_MasteredPeer{
		node:   node,
		wg:     xsync.NewWorkGroup(peerCtx),
		cancel: peerCancel,
		state0: m.node.State.Snapshot(), // XXX don't need .state0 if vvv is not moved to .acceptPeer
		// TODO change limiting by buffer size -> to limiting by time
		//      (see updateNodeTab for details)
		notifyq:         make(chan _ΔClusterState, 1024),
		notifyqOverflow: make(chan struct{}),
		acceptDone:      make(chan struct{}),
	}
	m.peerTab[node.NID] = peer

	// spawn task to send accept and proxy δnodeTab/δpartTab to the peer
	// XXX -> func m.acceptPeer ?
	peer.wg.Go(func(ctx context.Context) error {
		// go  main <- peer "peer (should be) disconnected"  when all peer's task finish
		m.mainWG.Go(func(ctx context.Context) error {
			// wait for all tasks related to peer to complete and then
			// notify main that peer node should go. Don't take ctx into
			// account - it is ~ runCtx and should be parent of context
			// under which per-peer tasks are spawned. This way if runCtx
			// is canceled -> any per-peer ctx should be canceled too and
			// wg.Wait should not block.
			err := peer.wg.Wait()

			select {
			case <-ctx.Done():
				return ctx.Err()
			case m.nodeLeaveq <- nodeLeave{peer, err}:
				return nil // ok
			}
		})

		// XXX err -> indicated that accept0 failed ?

		// XXX close link on ctx cancel?
		link := peer.node.Link()

		// send acceptance to just identified peer
		err := n.req.Reply(accept)
		if err != nil {
			return fmt.Errorf("send accept: %w", err)
		}

		// send initial state snapshot to accepted node

		// nodeTab
		err = link.Send1(&peer.state0.NodeTab)
		if err != nil {
			return fmt.Errorf("send nodeTab: %w", err)
		}

		// partTab (not to S until cluster is RUNNING)
		if !(peer.node.Type == proto.STORAGE && peer.state0.Code != proto.ClusterRunning) {
			err = link.Send1(&peer.state0.PartTab)
			if err != nil {
				return fmt.Errorf("send partTab: %w", err)
			}
		}

		// XXX send clusterState too? (NEO/py does not send it)

		// indicate to run that initial acceptance is done
		close(peer.acceptDone)

		// proxy δnodeTab,δpartTab/δclusterState from main to the peer
		return peer.notify(ctx)
	})

	return peer, true
}

// XXX run runs f after initial phase of peer acceptance is over.
//
// XXX this is very similar if a separate Accept call would return peers
// already identified and answered with initial accept message sequence.
// However identification needs decisions from main task (to e.g. consult
// nodeTab to see if peer laddr is not overlapping with anyone's, and to assign
// nid). Because main is involved we cannot move it to completely separate task
// and give main only one Accept entry point to call.
func (p *_MasteredPeer) run(ctx context.Context, f func() error) error {
	select {
	case <-ctx.Done():
		return ctx.Err()
	// XXX in general we should also wait for if "accept0 failed". However
	// as that means accept0 task error, it would cancel ctx for all other
	// tasks run through p.wg . And run is called with contexts whose
	// cancel is derived from wg cancel - so we don't check for that. XXX
	case <-p.acceptDone:
		return f()
	}
}

// notify proxies δnodeTab/δpeerTab/δClusterState update to the peer.
// XXX merge into m.acceptPeer ?
func (p *_MasteredPeer) notify(ctx context.Context) (err error) {
	defer task.Runningf(&ctx, "notify %s", p.node.NID)(&err)

	stateCode := p.state0.Code

	// XXX vvv right?
	return xcontext.WithCloseOnErrCancel(ctx, p.node.Link(), func() error {
		for {
			var δstate _ΔClusterState

			select {
			case <-ctx.Done():
				return ctx.Err()

			case <-p.notifyqOverflow:
				// XXX err -> ?
				return fmt.Errorf("detaching (peer is too slow to consume updates)")

			case δstate = <-p.notifyq:	// XXX could be also closed?
			}

			var msg proto.Msg
			switch δstate := δstate.(type) {
			case *_ΔNodeTab:
				msg = &proto.NotifyNodeInformation{
					IdTime:   proto.IdTimeNone, // FIXME
					NodeList: []proto.NodeInfo{δstate.NodeInfo},
				}

			case *_ΔPartTab:
				// don't send δpartTab to S unless cluster is RUNNING
				if p.node.Type == proto.STORAGE && stateCode != proto.ClusterRunning {
					continue
				}
				msg = &proto.NotifyPartitionChanges{
					// XXX
					// PTid: ...,
					// NumReplicas: ...,
					// CellList: ...,
				}
				panic("TODO") // ^^^

			case *_ΔStateCode:
				stateCode = δstate.ClusterState
				msg = &proto.NotifyClusterState{
					State: stateCode,
				}

			default:
				panic("bug")
			}

			err := p.node.Link().Send1(msg)
			if err != nil {
				return err
			}
		}
	})

	return nil
}


// allocNID allocates new node ID for a node of kind nodeType.
// XXX it is bad idea for master to assign node ID to coming node
// -> better nodes generate really unique UUID themselves and always show with them
func (m *Master) allocNID(nodeType proto.NodeType) proto.NodeID {
	for num := int32(1); num < 1<<24; num++ {
		nid := proto.NID(nodeType, num)
		if m.node.State.NodeTab.Get(nid) == nil {
			return nid
		}
	}

	panic("all nid allocated ???")	// XXX more robust ?
}
